"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.commands = void 0;
const lodash_1 = __importDefault(require("lodash"));
const path_1 = __importDefault(require("path"));
const driver_1 = require("appium/driver");
const support_1 = require("appium/support");
const constants_1 = require("../constants");
// List of env variables, that can be expanded in path
const KNOWN_ENV_VARS = [
    'APPDATA', 'LOCALAPPDATA',
    'PROGRAMFILES', 'PROGRAMFILES(X86)',
    'PROGRAMDATA', 'ALLUSERSPROFILE',
    'TEMP', 'TMP',
    'HOMEPATH', 'USERPROFILE', 'PUBLIC'
];
const commands = {};
exports.commands = commands;
commands.pushFile = async function pushFile(remotePath, base64Data) {
    this.ensureFeatureEnabled(constants_1.MODIFY_FS_FEATURE);
    if (remotePath.endsWith(path_1.default.sep)) {
        throw new driver_1.errors.InvalidArgumentError('It is expected that remote path points to a file rather than a folder. ' +
            `'${remotePath}' is given instead`);
    }
    if (lodash_1.default.isArray(base64Data)) {
        // some clients (ahem) java, send a byte array encoding utf8 characters
        // instead of a string, which would be infinitely better!
        base64Data = Buffer.from(base64Data).toString('utf8');
    }
    const fullPath = resolveToAbsolutePath(remotePath);
    await (0, support_1.mkdirp)(path_1.default.dirname(fullPath));
    const content = Buffer.from(base64Data, 'base64');
    await support_1.fs.writeFile(fullPath, content);
};
commands.pullFile = async function pullFile(remotePath) {
    const fullPath = resolveToAbsolutePath(remotePath);
    await checkFileExists(fullPath);
    return (await support_1.util.toInMemoryBase64(fullPath)).toString();
};
commands.pullFolder = async function pullFolder(remotePath) {
    const fullPath = resolveToAbsolutePath(remotePath);
    await checkFolderExists(fullPath);
    return (await support_1.zip.toInMemoryZip(fullPath, {
        encodeToBase64: true,
    })).toString();
};
/**
 * @typedef {Object} DeleteFileOptions
 * @property {string} remotePath - The path to a file.
 * The path may contain environment variables that could be expanded on the server side.
 * Due to security reasons only variables listed below would be expanded: `APPDATA`,
 * `LOCALAPPDATA`, `PROGRAMFILES`, `PROGRAMFILES(X86)`, `PROGRAMDATA`, `ALLUSERSPROFILE`,
 * `TEMP`, `TMP`, `HOMEPATH`, `USERPROFILE`, `PUBLIC`.
 */
/**
 * Remove the file from the file system
 *
 * @param {DeleteFileOptions} opts
 * @throws {InvalidArgumentError} If the file to be deleted does not exist or
 * remote path is not an absolute path.
 */
commands.windowsDeleteFile = async function windowsDeleteFile(opts) {
    this.ensureFeatureEnabled(constants_1.MODIFY_FS_FEATURE);
    const { remotePath } = opts ?? {};
    const fullPath = resolveToAbsolutePath(remotePath);
    await checkFileExists(fullPath);
    await support_1.fs.unlink(fullPath);
};
/**
 * @typedef {Object} DeleteFolderOptions
 * @property {string} remotePath - The path to a folder.
 * The path may contain environment variables that could be expanded on the server side.
 * Due to security reasons only variables listed below would be expanded: `APPDATA`,
 * `LOCALAPPDATA`, `PROGRAMFILES`, `PROGRAMFILES(X86)`, `PROGRAMDATA`, `ALLUSERSPROFILE`,
 * `TEMP`, `TMP`, `HOMEPATH`, `USERPROFILE`, `PUBLIC`.
 */
/**
 * Remove the folder from the file system
 *
 * @param {DeleteFolderOptions} opts
 * @throws {InvalidArgumentError} If the folder to be deleted does not exist or
 * remote path is not an absolute path.
 */
commands.windowsDeleteFolder = async function windowsDeleteFolder(opts) {
    this.ensureFeatureEnabled(constants_1.MODIFY_FS_FEATURE);
    const { remotePath } = opts ?? {};
    const fullPath = resolveToAbsolutePath(remotePath);
    await checkFolderExists(fullPath);
    await support_1.fs.rimraf(fullPath);
};
function resolveToAbsolutePath(remotePath) {
    const resolvedPath = remotePath.replace(/%([^%]+)%/g, (_, key) => KNOWN_ENV_VARS.includes(key.toUpperCase())
        ? process.env[key.toUpperCase()]
        : `%${key}%`);
    if (!path_1.default.isAbsolute(resolvedPath)) {
        throw new driver_1.errors.InvalidArgumentError('It is expected that remote path is absolute. ' +
            `'${resolvedPath}' is given instead`);
    }
    return resolvedPath;
}
async function checkFileExists(remotePath) {
    if (!await support_1.fs.exists(remotePath)) {
        throw new driver_1.errors.InvalidArgumentError(`The remote file '${remotePath}' does not exist.`);
    }
    const stat = await support_1.fs.stat(remotePath);
    if (!stat.isFile()) {
        throw new driver_1.errors.InvalidArgumentError('It is expected that remote path points to a file rather than a folder. ' +
            `'${remotePath}' is given instead`);
    }
}
async function checkFolderExists(remotePath) {
    if (!await support_1.fs.exists(remotePath)) {
        throw new driver_1.errors.InvalidArgumentError(`The remote folder '${remotePath}' does not exist.`);
    }
    const stat = await support_1.fs.stat(remotePath);
    if (!stat.isDirectory()) {
        throw new driver_1.errors.InvalidArgumentError('It is expected that remote path points to a folder rather than a file. ' +
            `'${remotePath}' is given instead`);
    }
}
exports.default = commands;
//# sourceMappingURL=file-movement.js.map