export default commands;
export type ClickOptions = {
    /**
     * Hexadecimal identifier of the element to click on.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     */
    elementId: string;
    /**
     * Integer horizontal coordinate of the click point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    x: number;
    /**
     * Integer vertical coordinate of the click point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    y: number;
    /**
     * [left] Name of
     * the mouse button to be clicked. An exception is thrown if an unknown button name
     * is provided.
     */
    button: "left" | "middle" | "right" | "back" | "forward";
    /**
     * List of possible keys or a single key name to
     * depress while the click is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
     * ['ctrl', 'alt']
     */
    modifierKeys: string[] | string;
    /**
     * The number of milliseconds to wait between pressing
     * and releasing the mouse button. By default no delay is applied, which simulates a
     * regular click.
     */
    durationMs: number;
    /**
     * [1] How many times the click must be performed.
     */
    times: number;
    /**
     * [100] Duration od the pause between each
     * click gesture. Only makes sense if `times` is greater than one.
     */
    interClickDelayMs: number;
};
export type ScrollOptions = {
    /**
     * Hexadecimal identifier of the element to scroll.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     */
    elementId: string;
    /**
     * Integer horizontal coordinate of the scroll point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    x: number;
    /**
     * Integer vertical coordinate of the scroll point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    y: number;
    /**
     * Integer horizontal scroll delta. Either this value
     * or deltaY must be provided, but not both.
     */
    deltaX: number;
    /**
     * Integer vertical scroll delta. Either this value
     * or deltaX must be provided, but not both.
     */
    deltaY: number;
    /**
     * List of possible keys or a single key name to
     * depress while the scroll is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
     * ['ctrl', 'alt']
     */
    modifierKeys: string[] | string;
};
export type ClickAndDragOptions = {
    /**
     * Hexadecimal identifier of the element to start the drag from.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     */
    startElementId: string;
    /**
     * Integer horizontal coordinate of the drag start point. Both startX
     * and startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    startX: number;
    /**
     * Integer vertical coordinate of the drag start point. Both startX and
     * startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    startY: number;
    /**
     * Hexadecimal identifier of the element to end the drag on.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     */
    endElementId: string;
    /**
     * Integer horizontal coordinate of the drag end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    endX: number;
    /**
     * Integer vertical coordinate of the drag end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    endY: number;
    /**
     * List of possible keys or a single key name to
     * depress while the drag is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
     * ['ctrl', 'alt']
     */
    modifierKeys: string[] | string;
    /**
     * [5000] The number of milliseconds to wait between pressing
     * the left mouse button and moving the cursor to the ending drag point.
     */
    durationMs: number;
};
export type HoverOptions = {
    /**
     * Hexadecimal identifier of the element to start the hover from.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     */
    startElementId: string;
    /**
     * Integer horizontal coordinate of the hover start point. Both startX
     * and startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    startX: number;
    /**
     * Integer vertical coordinate of the hover start point. Both startX and
     * startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    startY: number;
    /**
     * Hexadecimal identifier of the element to end the hover on.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     */
    endElementId: string;
    /**
     * Integer horizontal coordinate of the hover end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    endX: number;
    /**
     * Integer vertical coordinate of the hover end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     */
    endY: number;
    /**
     * List of possible keys or a single key name to
     * depress while the hover is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while hovering, provide the value of
     * ['ctrl', 'alt']
     */
    modifierKeys: string[] | string;
    /**
     * [500] The number of milliseconds between
     * moving the cursor from the starting to the ending hover point.
     */
    durationMs: number;
};
export type KeyAction = {
    /**
     * Allows to set a delay in milliseconds between key input series.
     * Either this property or `text` or `virtualKeyCode` must be provided.
     */
    pause: number;
    /**
     * Non-empty string of Unicode text to type.
     * Either this property or `pause` or `virtualKeyCode` must be provided.
     */
    text: string;
    /**
     * Valid virtual key code. The list of supported key codes
     * is available at https://learn.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes
     * Either this property or `pause` or `text` must be provided.
     */
    virtualKeyCode: number;
    /**
     * [undefined] This property only makes sense in combination with `virtualKeyCode`.
     * If set to `true` then the corresponding key will be depressed, `false` - released. By default
     * the key is just pressed once.
     * ! Do not forget to release depressed keys in your automated tests.
     */
    down: boolean | null;
};
export type KeysOptions = {
    /**
     * One or more key actions.
     */
    actions: KeyAction[] | KeyAction;
};
export namespace commands {
    /**
     * @typedef {Object} ClickOptions
     * @property {string} elementId Hexadecimal identifier of the element to click on.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     * @property {number} x Integer horizontal coordinate of the click point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {number} y Integer vertical coordinate of the click point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {'left' | 'middle' | 'right' | 'back' | 'forward'} button [left] Name of
     * the mouse button to be clicked. An exception is thrown if an unknown button name
     * is provided.
     * @property {string[]|string} modifierKeys List of possible keys or a single key name to
     * depress while the click is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
     * ['ctrl', 'alt']
     * @property {number} durationMs The number of milliseconds to wait between pressing
     * and releasing the mouse button. By default no delay is applied, which simulates a
     * regular click.
     * @property {number} times [1] How many times the click must be performed.
     * @property {number} interClickDelayMs [100] Duration od the pause between each
     * click gesture. Only makes sense if `times` is greater than one.
     */
    /**
     * Performs single click mouse gesture.
     *
     * @param {ClickOptions} opts
     * @throws {Error} If given options are not acceptable or the gesture has failed.
     */
    function windowsClick(opts: ClickOptions): Promise<void>;
    /**
     * @typedef {Object} ScrollOptions
     * @property {string} elementId Hexadecimal identifier of the element to scroll.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     * @property {number} x Integer horizontal coordinate of the scroll point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {number} y Integer vertical coordinate of the scroll point. Both x and y coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {number} deltaX Integer horizontal scroll delta. Either this value
     * or deltaY must be provided, but not both.
     * @property {number} deltaY Integer vertical scroll delta. Either this value
     * or deltaX must be provided, but not both.
     * @property {string[]|string} modifierKeys List of possible keys or a single key name to
     * depress while the scroll is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
     * ['ctrl', 'alt']
     */
    /**
     * Performs horizontal or vertical scrolling with mouse wheel.
     *
     * @param {ScrollOptions} opts
     * @throws {Error} If given options are not acceptable or the gesture has failed.
     */
    function windowsScroll(opts: ScrollOptions): Promise<void>;
    /**
     * @typedef {Object} ClickAndDragOptions
     * @property {string} startElementId Hexadecimal identifier of the element to start the drag from.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     * @property {number} startX Integer horizontal coordinate of the drag start point. Both startX
     * and startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {number} startY Integer vertical coordinate of the drag start point. Both startX and
     * startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {string} endElementId Hexadecimal identifier of the element to end the drag on.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     * @property {number} endX Integer horizontal coordinate of the drag end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {number} endY Integer vertical coordinate of the drag end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {string[]|string} modifierKeys List of possible keys or a single key name to
     * depress while the drag is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while clicking, provide the value of
     * ['ctrl', 'alt']
     * @property {number} durationMs [5000] The number of milliseconds to wait between pressing
     * the left mouse button and moving the cursor to the ending drag point.
     */
    /**
     * Performs drag and drop mouse gesture.
     *
     * @param {ClickAndDragOptions} opts
     * @throws {Error} If given options are not acceptable or the gesture has failed.
     */
    function windowsClickAndDrag(opts: ClickAndDragOptions): Promise<void>;
    /**
     * @typedef {Object} HoverOptions
     * @property {string} startElementId Hexadecimal identifier of the element to start the hover from.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     * @property {number} startX Integer horizontal coordinate of the hover start point. Both startX
     * and startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {number} startY Integer vertical coordinate of the hover start point. Both startX and
     * startY coordinates must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {string} endElementId Hexadecimal identifier of the element to end the hover on.
     * If this parameter is missing then given coordinates will be parsed as absolute ones.
     * Otherwise they are parsed as relative to the top left corner of this element.
     * @property {number} endX Integer horizontal coordinate of the hover end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {number} endY Integer vertical coordinate of the hover end point. Both endX and endY coordinates
     * must be provided or none of them if elementId is present. In such case the gesture
     * will be performed at the center point of the given element.
     * @property {string[]|string} modifierKeys List of possible keys or a single key name to
     * depress while the hover is being performed. Supported key names are: Shift, Ctrl, Alt, Win.
     * For example, in order to keep Ctrl+Alt depressed while hovering, provide the value of
     * ['ctrl', 'alt']
     * @property {number} durationMs [500] The number of milliseconds between
     * moving the cursor from the starting to the ending hover point.
     */
    /**
     * Performs hover mouse gesture.
     *
     * @param {HoverOptions} opts
     * @throws {Error} If given options are not acceptable or the gesture has failed.
     */
    function windowsHover(opts: HoverOptions): Promise<void>;
    /**
     * @typedef {Object} KeyAction
     * @property {number} pause Allows to set a delay in milliseconds between key input series.
     * Either this property or `text` or `virtualKeyCode` must be provided.
     * @property {string} text Non-empty string of Unicode text to type.
     * Either this property or `pause` or `virtualKeyCode` must be provided.
     * @property {number} virtualKeyCode Valid virtual key code. The list of supported key codes
     * is available at https://learn.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes
     * Either this property or `pause` or `text` must be provided.
     * @property {boolean?} down [undefined] This property only makes sense in combination with `virtualKeyCode`.
     * If set to `true` then the corresponding key will be depressed, `false` - released. By default
     * the key is just pressed once.
     * ! Do not forget to release depressed keys in your automated tests.
     */
    /**
     * @typedef {Object} KeysOptions
     * @property {KeyAction[] | KeyAction} actions One or more key actions.
     */
    /**
     * Performs customized keyboard input.
     *
     * @param {KeysOptions} opts
     * @throws {Error} If given options are not acceptable or the gesture has failed.
     */
    function windowsKeys(opts: KeysOptions): Promise<void>;
}
//# sourceMappingURL=gestures.d.ts.map