export function createKeyInput(params?: {}): {
    type: number;
    union: {
        ki: {
            wVk: number;
            time: number;
            dwExtraInfo: number;
            wScan: number;
            dwFlags: number;
        };
    };
};
/**
 * Sends the provided input structures to SendInput WinAPI
 *
 * @param {object|object[]} inputs single INPUT structure or
 * an array of input structures. Consider combining keyboard inputs only.
 * @throws {Error} If any of the given inputs has not been successfully executed.
 */
export function handleInputs(inputs: object | object[]): Promise<any>;
/**
 * @typedef {Object} MouseButtonOptions
 * @property {typeof MOUSE_BUTTON[keyof typeof MOUSE_BUTTON]} button The desired button name to click
 * @property {typeof MOUSE_BUTTON_ACTION[keyof typeof MOUSE_BUTTON_ACTION]} action The desired button action
 */
/**
 * Transforms given mouse button parameters into an appropriate
 * input structure
 *
 * @param {MouseButtonOptions} opts
 * @returns {Promise<INPUT>} The resulting input structure
 * @throws {Error} If the input data is invalid
 */
export function toMouseButtonInput({ button, action }: MouseButtonOptions): Promise<any>;
/**
 * @typedef {Object} MouseMoveOptions
 * @property {number} dx Horizontal delta relative to the current cursor position as an integer.
 * Most be provided if dy is present
 * @property {number} dy Vertical delta relative to the current cursor position as an integer.
 * Most be provided if dx is present
 * @property {number} x Horizontal absolute cursor position on the virtual desktop as an integer.
 * Most be provided if y is present
 * @property {number} y Vertical absolute cursor position on the virtual desktop as an integer.
 * Most be provided if x is present
 */
/**
 * Transforms given mouse move parameters into an appropriate
 * input structure
 *
 * @param {Partial<MouseMoveOptions>} opts
 * @param {Size?} screenSize
 * @returns {Promise<INPUT>} The resulting input structure
 * @throws {Error} If the input data is invalid
 */
export function toMouseMoveInput({ dx, dy, x, y }: Partial<MouseMoveOptions>, screenSize?: Size | null): Promise<any>;
/**
 * @typedef {Object} MouseWheelOptions
 * @property {number} dx Horizontal scroll delta as an integer.
 * If provided then no vertical scroll delta must be set.
 * @property {number} dy Vertical scroll delta as an integer.
 * If provided then no horizontal scroll delta must be set.
 */
/**
 * Transforms given mouse wheel parameters into an appropriate
 * input structure
 *
 * @param {MouseWheelOptions} opts
 * @returns {INPUT | null} The resulting input structure or null
 * if no input has been generated.
 * @throws {Error} If the input data is invalid
 */
export function toMouseWheelInput({ dx, dy }: MouseWheelOptions): any | null;
/**
 * Transforms the given Unicode text into an array of inputs
 * ready to be used as parameters for SendInput API
 *
 * @param {string} text An arbitrary Unicode string
 * @returns {object[]} Array of key inputs
 */
export function toUnicodeKeyInputs(text: string): object[];
/**
 * @typedef {Object} Size
 * @property {number} width
 * @property {number} height
 */
/**
 * Fetches the size of the virtual screen
 *
 * @returns {Promise<Size>}
 */
export function getVirtualScreenSize(): Promise<Size>;
export const KEYEVENTF_KEYUP: 2;
export const KEY_ACTION: Readonly<{
    UP: "up";
    DOWN: "down";
}>;
export const MODIFIER_KEY: Readonly<{
    shift: 16;
    ctrl: 17;
    alt: 18;
    win: 91;
}>;
export const MOUSE_BUTTON: Readonly<{
    LEFT: "left";
    MIDDLE: "middle";
    RIGHT: "right";
    BACK: "back";
    FORWARD: "forward";
}>;
export const MOUSE_BUTTON_ACTION: Readonly<{
    UP: "up";
    DOWN: "down";
    CLICK: "click";
}>;
export type MouseButtonOptions = {
    /**
     * The desired button name to click
     */
    button: (typeof MOUSE_BUTTON)[keyof typeof MOUSE_BUTTON];
    /**
     * The desired button action
     */
    action: (typeof MOUSE_BUTTON_ACTION)[keyof typeof MOUSE_BUTTON_ACTION];
};
export type MouseMoveOptions = {
    /**
     * Horizontal delta relative to the current cursor position as an integer.
     * Most be provided if dy is present
     */
    dx: number;
    /**
     * Vertical delta relative to the current cursor position as an integer.
     * Most be provided if dx is present
     */
    dy: number;
    /**
     * Horizontal absolute cursor position on the virtual desktop as an integer.
     * Most be provided if y is present
     */
    x: number;
    /**
     * Vertical absolute cursor position on the virtual desktop as an integer.
     * Most be provided if x is present
     */
    y: number;
};
export type MouseWheelOptions = {
    /**
     * Horizontal scroll delta as an integer.
     * If provided then no vertical scroll delta must be set.
     */
    dx: number;
    /**
     * Vertical scroll delta as an integer.
     * If provided then no horizontal scroll delta must be set.
     */
    dy: number;
};
export type Size = {
    width: number;
    height: number;
};
//# sourceMappingURL=user32.d.ts.map